from gyver.database.drivers.dialect import DialectInfo
from gyver.database.typedef import Driver

from .interface import Dialect


def resolve_driver(driver: Driver) -> Dialect:
    """Resolve a driver to its associated DialectInfo.

    Args:
        driver (Driver): The database driver.

    Returns:
        Dialect: The DialectInfo associated with the provided driver.
    """
    _table: dict[Driver, Dialect] = {
        Driver.MYSQL: DialectInfo(
            default_port=3306,
            driver=Driver.MYSQL,
            dialect_name="mysql",
            async_driver="aiomysql",
            sync_driver="pymysql",
            only_host=False,
        ),
        Driver.POSTGRES: DialectInfo(
            default_port=5432,
            driver=Driver.POSTGRES,
            dialect_name="postgresql",
            async_driver="asyncpg",
            sync_driver="psycopg2",
            only_host=False,
        ),
        Driver.SQLITE: DialectInfo(
            default_port=0,
            driver=Driver.SQLITE,
            dialect_name="sqlite",
            async_driver="aiosqlite",
            sync_driver="",
            only_host=True,
        ),
        Driver.MARIADB: DialectInfo(
            default_port=3306,
            driver=Driver.MARIADB,
            dialect_name="mariadb",
            async_driver="aiomysql",
            sync_driver="pymysql",
            only_host=False,
        ),
    }
    return _table[driver]


def build_dialect_scheme(dialect: Dialect, sync: bool = False):
    """Build a dialect scheme string based on the Dialect.

    Args:
        dialect (Dialect): The Dialect object.
        sync (bool, optional): Whether to build a sync scheme. Defaults to False.

    Returns:
        str: The dialect scheme string.
    """
    return "+".join(
        item
        for item in (
            dialect.dialect_name,
            dialect.sync_driver if sync else dialect.async_driver,
        )
        if item
    )



from typing import Any
from typing import TypeVar

import typing_extensions
from gyver.attrs import define
from gyver.attrs.field import FieldInfo
from gyver.attrs.field import info

# Sentinel object used to check if the __config_class__ attribute
# has been set on the class
SENTINEL = object()

# Generic type variable T
T = TypeVar("T")


def mark(cls: type[T]) -> type[T]:
    """
    Mark a given class as a config class.
    If the attribute __config_class__ already exists,
    but is not the SENTINEL, it raises a TypeError indicating that
    the attribute has an unexpected value.
    It is used by the [resolve_confignames][gyver.config.AdapterConfigFactory] to
    differentiate config classes

    Args:
        cls (type[T]): The class to be marked as a config class.

    Returns:
        type[T]: The class, marked as a config class.
    """
    if getattr(cls, "__config_class__", SENTINEL) is not SENTINEL:
        raise TypeError("Variable __config_class__ has an unexpected value")
    setattr(cls, "__config_class__", SENTINEL)
    return cls


def is_config(obj: Any) -> bool:
    """
    Check if an object is marked as a config class.

    This function checks if the attribute __config_class__
    exists on the object and if it is equal to SENTINEL.

    Args:
        obj (Any): The object to be checked.

    Returns:
        bool: True if the object is marked as a config class, False otherwise.
    """
    return getattr(obj, "__config_class__", None) is SENTINEL


@typing_extensions.dataclass_transform(
    order_default=True,
    frozen_default=True,
    kw_only_default=False,
    field_specifiers=(FieldInfo, info),
)
def as_config(cls: type[T]) -> type[T]:
    """
    Transform a class into a config class.

    This function applies the @define decorator from gyver.attrs and also marks
    the class as a config_class

    Args:
        cls (type[T]): The class to be transformed into a config class.

    Returns:
        type[T]: The transformed class, marked as a config class.
    """
    return mark(define(cls))



import asyncio
from time import time
from typing import Any
from typing import Callable
from typing import Coroutine
from typing import Generic
from typing import Optional
from typing import TypeVar

from gyver.attrs import call_init
from gyver.attrs import mutable

from gyver.exc import ErrorGroup

from .resource import Resource

T = TypeVar("T")


FactoryType = Callable[[], Coroutine[Any, Any, T]]
ReleaserType = Callable[[T], Coroutine[Any, Any, None]]


@mutable
class AsyncPool(Generic[T]):
    resources: asyncio.Queue[Resource[T]]
    factory: FactoryType[T]
    releaser: ReleaserType[T]
    _pool_recycle: float
    _pool_size: int
    _available: int
    _available_semaphore: asyncio.Lock

    def __init__(
        self,
        factory: FactoryType[T],
        releaser: ReleaserType[T],
        queue_class: type[asyncio.Queue[T]] = asyncio.LifoQueue,
        pool_size: int = 10,
        pool_recycle: float = 3600,
    ):
        """
        Initialize the AsyncPool.

        :param factory: A coroutine function that creates a new resource.
        :param releaser: A coroutine function that releases a resource.
        :param queue_class: The class to use for the resources queue. Defaults to asyncio.LifoQueue.
        :param pool_size: The maximum size of the resource pool. Defaults to 10.
        :param pool_recycle: The time in seconds after which a resource is considered expired and should be recycled. Defaults to 3600 (1 hour).
        """
        call_init(
            self,
            resources=queue_class(pool_size),
            factory=factory,
            releaser=releaser,
            pool_recycle=pool_recycle,
            pool_size=pool_size,
            available=pool_size,
            available_semaphore=asyncio.Lock(),
        )

    async def _initialize_resource(self) -> Resource[T]:
        """Creates a resource from the factory and returns it wrapped in the Resource object
        :return: The Resource wrapper
        """
        resource = await self.factory()
        return Resource.from_now(resource)

    async def _maybe_recycle(
        self, resource: Resource[T], current_ts: Optional[float] = None
    ) -> T:
        """Returns resource from resource object or
        releases the current resource and returns a new one if
        resource is expired.

        :param resource: The resource wrapper to be "maybe" recycled.
        :param current_ts: timestamp, defaults to `time.time()`

        :return: The resource acquired.
        """
        current_ts = current_ts or time()
        if (
            self._pool_recycle > 0
            and resource.last_usage + self._pool_recycle <= current_ts
        ):
            await self.releaser(resource.get())
            resource = await self._initialize_resource()
        return resource.get()

    async def acquire(self) -> T:
        """
        Acquires a resource from the pool, waiting until available.

        :return: T
        """
        acquired = await self._decrease_available()
        if not acquired:
            resource = await self.resources.get()
        else:
            try:
                resource = self.resources.get_nowait()
            except asyncio.QueueEmpty:
                resource = await self._initialize_resource()
        return await self._maybe_recycle(resource)

    async def release(self, resource: T) -> None:
        """Puts resource back in queue and increases the
        amount of resources available to be acquired

        :param resource: The resource to be put.
        :return: None
        """
        self.resources.put_nowait(Resource.from_resource(resource))
        await self._increase_available()

    async def prefill(self, count: Optional[int] = None) -> None:
        """
        Prefills the queue by the amount passed.

        :param count: The amount of resources to initialize, up to the
            `pool_size`. Defaults to the pool_size if None.
        """
        count = count or self._pool_size
        count = min(count, self._pool_size)

        while self.resources.qsize() < count and await self._decrease_available():
            resource = await self._initialize_resource()
            self.resources.put_nowait(resource)

    async def dispose(self) -> None:
        """
        Remove and release all items of from the queue.

        :return: None
        :raises: ErrorGroup if any error happened while closing
        """
        errors = []
        while not self.resources.empty():
            resource = await self.resources.get()
            try:
                await self.releaser(resource.get())
            except Exception as e:
                errors.append(e)
            await self._decrease_available()
        if errors:
            raise ErrorGroup("Could not kill all resources", *errors)

    async def _decrease_available(self):
        """
        Decrease the count of available resources.

        This method should not be called directly. It is used internally
        by the `acquire` and `dispose` methods to keep track of the number
        of available resources.
        """
        async with self._available_semaphore:
            if self._available == 0:
                return False
            self._available -= 1
            return True

    async def _increase_available(self):
        """
        Increase the count of available resources.

        This method should not be called directly. It is used internally by the
        `release` and `prefill` methods to keep track of the number
        of available resources.
        """
        async with self._available_semaphore:
            if self._available == self._pool_size:
                return False
            self._available += 1
            return True



from collections import defaultdict
from typing import Mapping
from typing import Optional
from urllib.parse import parse_qs
from urllib.parse import quote

from gyver.attrs import mutable
from typing_extensions import Self

from gyver.url.encode import Encodable


@mutable(eq=False)
class Query(Encodable):
    """Represents a query string and provides methods to manipulate and encode it.

    Attributes:
        params (defaultdict[str, list[str]]): The dictionary containing query parameters.
    """

    params: defaultdict[str, list[str]]

    def __init__(self, querystr: str) -> None:
        """Initializes the Query object with the provided query string.

        Args:
            querystr (str): The input query string to parse.
        """
        self.params = defaultdict(list)
        self.params |= parse_qs(querystr, keep_blank_values=True)

    def encode(self) -> str:
        """Encodes the query object into a string representation.

        Returns:
            str: The encoded query string.
        """
        return "&".join(
            "=".join((quote(key, safe=""), quote(item or "", safe="")))
            for key, value in self.params.items()
            for item in value
        )

    def omit_empty_equal(self) -> str:
        """Encodes the query object into a string representation, omitting the equal sign if a value is empty.

        Returns:
            str: The encoded query string.
        """
        return "&".join(
            "=".join(map(quote, (key, item))) if item else quote(key)
            for key, value in self.params.items()
            for item in value
        )

    def add(self, args: Optional[Mapping[str, str]] = None, /, **params: str) -> Self:
        """Adds query parameters to the Query object.

        Args:
            args (Optional[Mapping[str, str]], optional): Additional query parameters as a mapping object.
            **params (str): Additional query parameters as keyword arguments.

        Returns:
            Query: The modified Query object.
        """
        query = {**(args or {}), **params}
        for key, value in query.items():
            self.params[key].append(value)
        return self

    def set(self, args: Optional[Mapping[str, str]] = None, /, **params: str) -> Self:
        """Sets the query parameters, replacing any existing parameters.

        Args:
            args (Optional[Mapping[str, str]], optional): New query parameters as a mapping object.
            **params (str): New query parameters as keyword arguments.

        Returns:
            Query: The modified Query object.
        """
        query = {**(args or {}), **params}
        self.params.clear()
        self.add(query)
        return self

    def __setitem__(self, key: str, value: str) -> None:
        """Sets a query parameter using the square bracket syntax.

        Args:
            key (str): The parameter key.
            value (str): The parameter value.
        """
        self.add({key: value})

    def __getitem__(self, key: str) -> list[str]:
        """Retrieves the values of a query parameter using the square bracket syntax.

        Args:
            key (str): The parameter key.

        Returns:
            list[str]: The list of parameter values.
        """
        return self.params[key]

    def remove(self, *keys: str) -> Self:
        """Removes query parameters from the Query object.

        Args:
            keys (str): The parameter keys to remove.

        Returns:
            Query: The modified Query object.
        """
        for key in keys:
            self.params.pop(key, None)
        return self

    def sort(self) -> Self:
        """Sorts the query parameters in lexicographic order based on the parameter names.

        Returns:
            Query: The modified Query object.
        """
        self.params = defaultdict(list, sorted(self.params.items()))
        return self



import re
from typing import Union
from urllib.parse import quote
from urllib.parse import unquote

from gyver.attrs import mutable
from typing_extensions import Self

from gyver.url.encode import Encodable
from gyver.url.utils import is_valid_encoded_path
from gyver.url.utils import utf8

PERCENT_REGEX = r"\%[a-fA-F\d][a-fA-F\d]"

SAFE_SEGMENT_CHARS = ":@-._~!$&'()*+,;="


@mutable(eq=False)
class Path(Encodable):
    """Represents a path in a URL.

    Attributes:
        segments (list[str]): List of path segments.
    """

    segments: list[str]

    def __init__(self, pathstr: str) -> None:
        """Initialize a Path object.

        Args:
            pathstr (str): The string representation of the path.
        """
        self.segments = self._load(pathstr)

    def _load(self, pathstr: str) -> list[str]:
        """Load the path segments from the path string.

        Args:
            pathstr (str): The string representation of the path.

        Returns:
            list[str]: List of path segments.
        """
        path_segments = pathstr.split("/")
        return self._process_segments([], path_segments)

    def _process_segments(
        self,
        source: list[str],
        to_append: list[str],
    ) -> list[str]:
        """Process the path segments.

        Args:
            source (list[str]): The source list of segments.
            to_append (list[str]): The list of segments to append.

        Returns:
            list[str]: The processed list of segments.
        """
        segments = list(source)
        for seg in to_append:
            if not is_valid_encoded_path.match(seg):
                seg = quote(utf8(seg))
            segments.append(seg)
        return [
            unquote(item.decode("utf-8") if isinstance(item, bytes) else item)
            for item in segments
        ]

    def encode(self) -> str:
        """Encode the path segments into a URL-encoded string.

        Returns:
            str: The URL-encoded path string.
        """
        return "/".join(quote(utf8(item), SAFE_SEGMENT_CHARS) for item in self.segments)

    def add(self, path: Union[str, "Path"]) -> Self:
        """Add path segments to the existing path.

        Args:
            path (Union[str, Path]): The path segments to add. Can be a string or a Path object.

        Returns:
            Path: The modified Path object.
        """
        segments = []
        if isinstance(path, str):
            segments.extend(self._process_segments([], path.lstrip("/").split("/")))
        else:
            segments.extend(path.segments)
        self.segments = self._process_segments(self.segments, segments)
        return self

    def set(self, path: Union[str, "Path"]) -> Self:
        """Set the path segments to a new value.

        Args:
            path (Union[str, Path]): The new path segments. Can be a string or a Path object.

        Returns:
            Path: The modified Path object.
        """
        segments = []
        if isinstance(path, str):
            segments.extend(self._process_segments([], path.split("/")))
        else:
            segments.extend(path.segments)
        self.segments = segments
        return self

    @property
    def isdir(self) -> bool:
        """Check if the path represents a directory.

        Path is considered a directory if it is empty or ends with a trailing slash.

        Returns:
            bool: True if the path is a directory, False otherwise.
        """
        return not self.segments or (self.segments[-1] == "")

    @property
    def isfile(self) -> bool:
        """Check if the path represents a file.

        Returns:
            bool: True if the path is a file, False if it is a directory.
        """
        return not self.isdir

    def normalize(self) -> Self:
        """Normalize the path by removing redundant segments.

        Returns:
            Path: The normalized Path object.
        """
        if resolved := self.encode():
            normalized = normalize(resolved)
            self.segments = self._load(normalized)
        return self


_duplicates_regex = re.compile(r"/+")


def normalize(path: str) -> str:
    """Normalize a path string by removing redundant segments.

    Args:
        path (str): The path string to normalize.

    Returns:
        str: The normalized path string.
    """
    stack = []
    path = _duplicates_regex.sub("/", path)
    for segment in path.split("/"):
        if segment == "..":
            if stack:
                stack.pop()
        elif segment not in [".", ""]:
            stack.append(segment)
        elif segment == ".":
            continue
    return "/" + "/".join(stack) + ("/" * path.endswith("/")) if stack else "/"



import typing

import sqlalchemy as sa

from .interface import BindClause
from .interface import Comparator
from .interface import Comparison
from .interface import FieldType
from .interface import Sortable


def always_true(field: FieldType, target: typing.Any) -> Comparison:
    del field, target
    return sa.true()


def equals(field: FieldType, target: typing.Any) -> Comparison:
    return field == target


def not_equals(field: FieldType, target: typing.Any) -> Comparison:
    return field != target


def greater(field: FieldType, target: Sortable) -> Comparison:
    return field > target


def greater_equals(field: FieldType, target: Sortable) -> Comparison:
    return field >= target


def lesser(field: FieldType, target: Sortable) -> Comparison:
    return field < target


def lesser_equals(field: FieldType, target: Sortable) -> Comparison:
    return field <= target


def between(field: FieldType, target: tuple[Sortable, Sortable]) -> Comparison:
    left, right = target
    return field.between(left, right)


def range(field: FieldType, target: tuple[Sortable, Sortable]) -> Comparison:
    left, right = target
    return sa.and_(greater_equals(field, left), (lesser(field, right)))


def like(field: FieldType, target: str) -> Comparison:
    return field.like(f"%{target}%")


def rlike(field: FieldType, target: str) -> Comparison:
    return field.like(f"{target}%")


def llike(field: FieldType, target: str) -> Comparison:
    return field.like(f"%{target}")


def insensitive_like(
    opt: typing.Literal["like", "rlike", "llike"] = "like"
) -> Comparator[str]:
    fmt = {"like": "%{target}%", "rlike": "%{target}", "llike": "{target}%"}[opt]

    def comparator(field: FieldType, target: str):
        return field.ilike(fmt.format(target=target))

    return comparator


def isnull(field: FieldType, target: bool) -> Comparison:
    return field.is_(None) if target else field.is_not(None)


def includes(field: FieldType, target: typing.Sequence) -> Comparison:
    return field.in_(list(target))


def excludes(field: FieldType, target: typing.Sequence) -> Comparison:
    return field.not_in(list(target))


def json_contains(field: FieldType, target: typing.Any) -> Comparison:
    return sa.func.json_contains(field, f'"{target}"')


def json_empty(field: FieldType, target: typing.Any) -> Comparison:
    func_length = sa.func.json_length(field)
    return func_length == 0 if target else func_length != 0


def make_relation_check(clause: BindClause) -> Comparator:
    def _relation_exists(field: FieldType, target: bool) -> Comparison:
        comp = clause.bind(field.class_)
        func = (
            field.has
            if field.property.direction.name.lower() not in ("onetomany", "manytomany")
            else field.any
        )
        result = func() if str(comp) == str(sa.true()) else func(comp)
        return result if target else ~result

    return _relation_exists



from config import MISSING
from config import Config
from config import EnvMapping
from config.enums import Env
from config.envconfig import DotFile
from config.envconfig import EnvConfig
from config.interface import ConfigLike
from config.utils import boolean_cast
from config.utils import joined_cast
from config.utils import valid_path
from config.utils import with_rule

from gyver.utils.strings import comma_separated

from .adapter import AdapterConfigFactory
from .adapter.mark import as_config
from .adapter.mark import mark

__all__ = [
    "boolean_cast",
    "comma_separated",
    "valid_path",
    "joined_cast",
    "with_rule",
    "Config",
    "EnvMapping",
    "MISSING",
    "AdapterConfigFactory",
    "mark",
    "as_config",
    "EnvConfig",
    "Env",
    "DotFile",
    "ConfigLike",
]



from typing import Protocol

from gyver.database.typedef import Driver


class Dialect(Protocol):
    """Protocol representing a database dialect.

    Attributes:
        default_port (int): The default port for the database dialect.
        driver (Driver): The driver associated with the dialect.
        dialect_name (str): The name of the database dialect.
        async_driver (str): The asynchronous driver name.
        sync_driver (str): The synchronous driver name.
        only_host (bool): Indicates if the dialect supports only host connections.
    """

    default_port: int
    driver: Driver
    dialect_name: str
    async_driver: str
    sync_driver: str
    only_host: bool



import contextlib
from contextlib import suppress
from dataclasses import is_dataclass
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import TypeVar
from typing import cast
from typing import get_args
from typing import get_origin

from config import MISSING
from config import Config
from config.exceptions import MissingName
from config.interface import ConfigLike
from config.utils import boolean_cast
from gyver.attrs import define
from gyver.attrs import mark_factory
from gyver.attrs.utils.functions import disassemble_type
from gyver.attrs.utils.typedef import DisassembledType
from pydantic import BaseModel
from pydantic import v1

from gyver.utils import finder
from gyver.utils import json
from gyver.utils import panic
from gyver.utils.strings import make_lex_separator

from .dataclass import DataclassResolverStrategy
from .gattrs import GyverAttrsResolverStrategy
from .interface import FieldResolverStrategy
from .mark import is_config
from .pydantic import PydanticResolverStrategy

_default_config = Config()

T = TypeVar("T")


def _try_each(*names: str, default: Any, cast: Any, config: ConfigLike):
    for name in names:
        with suppress(MissingName):
            return config(name, cast)
    if default is not MISSING:
        return default
    raise panic(MissingName, f"{', '.join(names)} not found and no default was given")


def _resolve_cast(outer_type: type) -> tuple[Any, bool]:
    _sequences = (list, tuple, set)
    origin = get_origin(outer_type)
    if outer_type is bool:
        return boolean_cast, False
    with contextlib.suppress(ValueError):
        AdapterConfigFactory.get_strategy_class(disassemble_type(outer_type))
        return outer_type, True
    if origin is None:
        return (
            make_lex_separator(outer_type) if outer_type in _sequences else outer_type
        ), False
    if (origin := get_origin(outer_type)) in _sequences:
        assert origin is not None
        args = get_args(outer_type)
        cast = args[0] if args else str
        return make_lex_separator(origin, cast), False
    return (
        _loads if dict in (origin, outer_type) else origin,  # type:ignore
        False,
    )


def _loads(val: Any) -> Any:
    return json.loads(val) if isinstance(val, str) else val


@define
class AdapterConfigFactory:
    """
    Factory for creating configuration instances based on model classes.
    """

    config: ConfigLike = _default_config

    @staticmethod
    def get_strategy_class(
        config_class: DisassembledType,
    ) -> type[FieldResolverStrategy]:
        """
        Get the appropriate strategy class for resolving fields in the configuration class.

        Args:
            config_class (type): The configuration class to resolve.

        Returns:
            type[FieldResolverStrategy]: The strategy class for resolving fields.
        """
        klass = config_class.origin or config_class.type_
        base_model_cls: list[type] = [BaseModel, v1.BaseModel]
        if hasattr(klass, "__gyver_attrs__"):
            return GyverAttrsResolverStrategy
        elif is_dataclass(klass):
            return DataclassResolverStrategy
        elif issubclass(klass, tuple(base_model_cls)):
            return PydanticResolverStrategy
        elif hasattr(klass, "__attrs_attrs__"):
            from .attrs import AttrsResolverStrategy

            return AttrsResolverStrategy
        raise ValueError("Unknown class definition")

    def load(
        self,
        model_cls: type[T],
        __prefix__: str = "",
        __sep__: str = "__",
        *,
        presets: Optional[dict[str, Any]] = None,
        **defaults: Any,
    ) -> T:
        """
        Load a configuration instance based on a model class.

        Args:
            model_cls (type[T]): The model class representing the configuration.
            __prefix__ (str): Optional prefix for configuration fields.
            __set__(str): Optional prefix to separate fields of nested models. Default is "__"
            presets (Optional[dict[str, Any]]): Optional preset values for fields.
            **defaults (Any): Default values for fields.

        Returns:
            T: The loaded configuration instance.
        """
        presets = presets or {}
        strategy_class = self.get_strategy_class(disassemble_type(model_cls))
        resolvers = tuple(
            resolver
            for field in strategy_class.iterfield(model_cls)
            if (resolver := strategy_class(field)).init_name() not in presets
        )
        result = {
            resolver.init_name(): self._get_value(
                model_cls,
                resolver,
                __prefix__,
                __sep__,
                defaults,
            )
            for resolver in resolvers
        }
        return model_cls(**result | presets)

    def maker(
        self,
        model_cls: type[T],
        __prefix__: str = "",
        __sep__: str = "__",
        *,
        presets: Optional[dict[str, Any]] = None,
        **defaults: Any,
    ) -> Callable[[], T]:
        """
        Create a factory function for loading configuration instances.

        Args:
            model_cls (type[T]): The model class representing the configuration.
            __prefix__ (str): Optional prefix for configuration fields.
            presets (Optional[dict[str, Any]]): Optional preset values for fields.
            **defaults (Any): Default values for fields.

        Returns:
            Callable[[], T]: The factory function for loading configuration instances.
        """

        @mark_factory
        def load():
            return self.load(
                model_cls, __prefix__, __sep__, presets=presets, **defaults
            )

        return load

    def _get_value(
        self,
        model_cls: type,
        resolver: FieldResolverStrategy,
        prefix: str,
        sep: str,
        defaults: Mapping[str, Any],
    ):
        names = self.resolve_names(model_cls, resolver, prefix)
        default = next(
            (result for name in names if (result := defaults.get(name))),
            resolver.default(),
        )
        cast, _is_config = _resolve_cast(resolver.cast())
        if _is_config:
            return self.load(
                cast,
                f"{next(iter(names))}{sep}",
                sep,
                defaults=defaults,
            )
        return _try_each(*names, default=default, cast=cast, config=self.config)

    def resolve_names(
        self, model_cls: type, resolver: FieldResolverStrategy, prefix: str
    ) -> Sequence[str]:
        names = resolver.names()
        prefix = prefix or getattr(model_cls, "__prefix__", "")
        without_prefix = cast(
            Sequence[str], getattr(model_cls, "__without_prefix__", ())
        )
        if not prefix or set(names).intersection(without_prefix):
            return (*names, *map(str.upper, names))

        processed_names = []
        for name in names:
            _prefix = f"{prefix}_" if not prefix.endswith("_") else prefix
            processed = f"{_prefix}{name}".lower()
            processed_names.extend((processed, processed.upper()))

        return tuple(processed_names)

    @classmethod
    def resolve_confignames(
        cls,
        root: Path,
    ) -> dict[type, tuple[Sequence[str], ...]]:
        """
        Resolve the environment variable names required by marked config classes.

        Args:
            root (Path): The root directory of the project.

        Returns:
            dict[type, tuple[Sequence[str], ...]]: A dictionary of config classes and their associated environment variable names.
        """
        builder = finder.FinderBuilder().add_validator(is_config).from_path(root)
        output = builder.find()
        tempself = cls()
        return {
            cfg: tuple(
                tempself.resolve_names(cfg, field, "")
                for field in cfg.__fields__.values()
            )
            for cfg in output.values()
        }



from . import json
from . import timezone
from .exc import panic
from .finder import FinderBuilder
from .finder import finder_builder
from .helpers import DeprecatedClass
from .helpers import cache
from .helpers import deprecated
from .helpers import frozen
from .helpers import merge_dicts
from .lazy import asynclazyfield
from .lazy import dellazy
from .lazy import lazyfield
from .lazy import setlazy
from .strings import to_camel
from .strings import to_snake
from .strings import upper_camel

__all__ = [
    "lazyfield",
    "asynclazyfield",
    "dellazy",
    "setlazy",
    "to_camel",
    "to_snake",
    "upper_camel",
    "json",
    "timezone",
    "frozen",
    "merge_dicts",
    "cache",
    "deprecated",
    "DeprecatedClass",
    "FinderBuilder",
    "finder_builder",
    "panic",
]



from .filetree import FileTree
from .typedef import File
from .typedef import Folder
from .typedef import TextFile
from .virtual import VirtualFileTree

__all__ = ["FileTree", "VirtualFileTree", "Folder", "File", "TextFile"]



from typing import Any

from lazyfields import lazy

from gyver import utils

try:
    from pydantic.v1 import BaseModel
except ImportError:
    from pydantic import BaseModel


class Model(BaseModel):  # type: ignore
    """
    Model is a BaseModel overload with opinions on JSON parsing mutability and alias generation.

    Configurations:
        - json_loads: Uses utils.json.loads for JSON loading.
        - json_dumps: Uses utils.json.dumps for JSON dumping.
        - frozen: Model instances are frozen by default.
        - orm_mode: Enables ORM mode for this model.
        - alias_generator: Uses utils.to_camel for alias generation.
        - allow_population_by_field_name: Allows population of fields by field name.
        - keep_untouched: Keeps attributes of type lazy untouched during population.
    """

    class Config:
        json_loads = utils.json.loads
        json_dumps = utils.json.dumps
        frozen = True
        orm_mode = True
        alias_generator = utils.to_camel
        allow_population_by_field_name = True
        keep_untouched = (lazy,)

    def __setattr__(self, name: str, value: Any):
        """
        Overloads the __setattr__ method to allow lazy fields to work correctly.

        Args:
            name (str): The name of the attribute.
            value (Any): The value to set for the attribute.
        """
        if isinstance(getattr(type(self), name, None), lazy):
            object.__setattr__(self, name, value)
        else:
            return super().__setattr__(name, value)


class MutableModel(Model):
    """
    A mutable version of the Model class that allows unfreezing of instances.

    Configurations:
        - frozen: Model instances are not frozen, allowing mutability.
    """

    class Config:
        frozen = False



from typing import Generic

from gyver.context.context import AsyncContext
from gyver.context.context import Context
from gyver.context.interfaces.adapter import AtomicAdapter
from gyver.context.interfaces.adapter import AtomicAsyncAdapter
from gyver.context.typedef import T


class BoundContext(Context[T], Generic[T]):
    """A context manager for managing atomic transactions with an adapter."""

    adapter: AtomicAdapter[T]

    def __init__(self, adapter: AtomicAdapter[T], context: Context[T]) -> None:
        """
        Initialize a BoundContext instance.

        Args:
            adapter (AtomicAdapter[T]): The atomic adapter for managing transactions.
            context (Context[T]): The underlying context to manage.
        """
        super().__init__(adapter)
        self._context = context

    def acquire(self):
        """
        Acquire the context for a transaction and begin an atomic operation.

        Returns:
            T: The acquired context.
        """
        with self._lock:
            client = self._context.acquire()
            self._stack += 1
            self.adapter.begin(client)
        return client

    def release(self, commit: bool = True):
        """
        Release the context, committing or rolling back the transaction if needed.

        Args:
            commit (bool, optional): Whether to commit the transaction. Defaults to True.
        """
        with self._lock:
            if self._stack == 1:
                if self.adapter.in_atomic(self._context.client):
                    if commit:
                        self.adapter.commit(self._context.client)
                    else:
                        self.adapter.rollback(self._context.client)
                self._context.release()
            self._stack -= 1

    def __exit__(self, *exc):
        """
        Exit the context manager and release the context.

        Args:
            *exc: Exception information.
        """
        self.release(not any(exc))


class AsyncBoundContext(AsyncContext[T], Generic[T]):
    """An asynchronous context manager for managing atomic transactions with an async adapter."""

    adapter: AtomicAsyncAdapter[T]

    def __init__(
        self, adapter: AtomicAsyncAdapter[T], context: AsyncContext[T]
    ) -> None:
        """
        Initialize an AsyncBoundContext instance.

        Args:
            adapter (AtomicAsyncAdapter[T]): The async atomic adapter for managing transactions.
            context (AsyncContext[T]): The underlying async context to manage.
        """
        super().__init__(adapter)
        self._context = context

    async def acquire(self):
        """
        Acquire the async context for a transaction and begin an atomic operation.

        Returns:
            T: The acquired async context.
        """
        async with self._lock:
            client = await self._context.acquire()
            if self._stack == 0:
                await self.adapter.begin(client)
            self._stack += 1
            return client

    async def release(self, commit: bool = True):
        """
        Release the async context, committing or rolling back the transaction if needed.

        Args:
            commit (bool, optional): Whether to commit the transaction. Defaults to True.
        """
        async with self._lock:
            client = await self._context.client()
            if self._stack == 1:
                if commit:
                    await self.adapter.commit(client)
                else:
                    await self.adapter.rollback(client)
            self._stack -= 1
        await self._context.release()

    async def __aexit__(self, *exc):
        """
        Exit the async context manager and release the async context.

        Args:
            *exc: Exception information.
        """
        await self.release(not any(exc))



from typing import Optional

from cryptography.fernet import Fernet
from cryptography.fernet import MultiFernet
from lazyfields import lazyfield

from gyver.config import AdapterConfigFactory

from .config import CryptoConfig


class FernetCryptoProvider:
    def __init__(self, config: Optional[CryptoConfig] = None) -> None:
        self._config = config or AdapterConfigFactory().load(CryptoConfig)

    @property
    def config(self):
        return self._config

    @lazyfield
    def _multi(self):
        f1 = Fernet(self._config.secret.encode("utf8"))
        fn = [Fernet(item.encode("utf8")) for item in self._config.spares]
        return MultiFernet([f1, *fn])

    def encrypt(self, secret: str) -> str:
        return self._multi.encrypt(secret.encode("utf8")).decode("utf8")

    def decrypt(self, digest: str) -> str:
        return self._multi.decrypt(digest.encode("utf8")).decode("utf8")



import lazyfields

from gyver.utils.helpers import deprecated

lazy = lazyfields.lazy
lazyfield = deprecated(lazyfields.lazyfield)
asynclazyfield = deprecated(lazyfields.asynclazyfield)
setlazy = deprecated(lazyfields.setlazy)
dellazy = deprecated(lazyfields.asynclazyfield)
force_set = deprecated(lazyfields.force_set)
force_del = deprecated(lazyfields.force_del)
is_initialized = deprecated(lazyfields.is_initialized)

__all__ = [
    "lazy",
    "lazyfields",
    "asynclazyfield",
    "setlazy",
    "dellazy",
    "force_set",
    "force_del",
    "is_initialized",
]



import asyncio
import typing
from contextlib import asynccontextmanager
from contextlib import contextmanager
from contextlib import suppress
from datetime import datetime
from datetime import timezone
from queue import Queue

from gyver.attrs import define
from gyver.attrs import info
from lazyfields import lazyfield

from gyver.exc import CacheMiss
from gyver.utils import json

from .interface import AsyncCacheInterface
from .interface import CacheInterface

T = typing.TypeVar("T")


@define
class MockAsyncCache(AsyncCacheInterface):
    key_map: typing.MutableMapping[str, typing.Any] = info(default_factory=dict)

    @lazyfield
    def _queue(self) -> asyncio.Queue[object]:
        size = 10
        queue = asyncio.Queue(size)
        for _ in range(size):
            queue.put_nowait(object())
        return queue

    @asynccontextmanager
    async def _in_queue(self):
        try:
            await self._queue.get()
            yield
        finally:
            await self._queue.put(object())

    async def get(
        self, name: str, cast: typing.Callable[[typing.Any], T] = json.loads
    ) -> T:
        async with self._in_queue():
            try:
                response, exp = self.key_map[name]
            except KeyError:
                raise CacheMiss from None
            else:
                if exp is not None and exp <= datetime.now(timezone.utc):
                    raise CacheMiss
                return cast(response)

    async def set(
        self,
        name: str,
        value: typing.Any,
        expires_at: typing.Union[datetime, typing.Literal[-1]],
        dumps: typing.Callable[[typing.Any], str] = json.dumps,
    ) -> None:
        async with self._in_queue():
            self.key_map[name] = (
                dumps(value),
                expires_at if expires_at != -1 else None,
            )

    async def map_get(
        self,
        map_name: str,
        name: str,
        cast: typing.Callable[[typing.Any], T] = json.loads,
    ) -> T:
        async with self._in_queue():
            try:
                val = self.key_map[map_name][name]
            except KeyError:
                raise CacheMiss from None
            else:
                return cast(val)

    async def map_set(
        self,
        map_name: str,
        name: str,
        value: typing.Any,
        dumps: typing.Callable[[typing.Any], str] = json.dumps,
    ) -> None:
        async with self._in_queue():
            if map_name not in self.key_map:
                self.key_map[map_name] = {}
            self.key_map[map_name][name] = dumps(value)

    async def delete(self, name: str, map_name: typing.Optional[str] = None) -> None:
        async with self._in_queue():
            used_map = self.key_map
            if map_name is not None:
                if map_name not in self.key_map:
                    return
                used_map = self.key_map[map_name]

            with suppress(KeyError):
                del used_map[name]


_KeyMapType = typing.MutableMapping[
    str,
    typing.Union[
        tuple[typing.Any, typing.Optional[datetime]],
        dict[str, typing.Any],
    ],
]


@define
class MockCache(CacheInterface):
    key_map: _KeyMapType = info(default_factory=info)

    @lazyfield
    def _queue(self) -> Queue[object]:
        size = 10
        queue = Queue(size)
        for _ in range(size):
            queue.put_nowait(object())
        return queue

    @contextmanager
    def _in_queue(self):
        try:
            self._queue.get()
            yield
        finally:
            self._queue.put(object())

    def get(self, name: str, cast: typing.Callable[[typing.Any], T] = json.loads) -> T:
        with self._in_queue():
            try:
                val = self.key_map[name]
                if isinstance(val, dict):
                    raise CacheMiss
                response, exp = val
            except KeyError:
                raise CacheMiss from None
            else:
                if exp is not None and exp <= datetime.now(timezone.utc):
                    raise CacheMiss
                return cast(response)

    def set(
        self,
        name: str,
        value: typing.Any,
        expires_at: typing.Union[datetime, typing.Literal[-1]],
        dumps: typing.Callable[[typing.Any], str] = json.dumps,
    ) -> None:
        with self._in_queue():
            self.key_map[name] = (
                dumps(value),
                expires_at if expires_at != -1 else None,
            )

    def map_get(
        self,
        map_name: str,
        name: str,
        cast: typing.Callable[[typing.Any], T] = json.loads,
    ) -> T:
        with self._in_queue():
            try:
                mapping = self.key_map[map_name]
                if not isinstance(mapping, dict):
                    raise CacheMiss
                val = mapping[name]
            except KeyError:
                raise CacheMiss from None
            else:
                return cast(val)

    def map_set(
        self,
        map_name: str,
        name: str,
        value: typing.Any,
        dumps: typing.Callable[[typing.Any], str] = json.dumps,
    ) -> None:
        with self._in_queue():
            if map_name not in self.key_map:
                self.key_map[map_name] = {}

            if isinstance(self.key_map[map_name], tuple):
                self.key_map[map_name] = {}
            mapping = typing.cast(dict, self.key_map[map_name])
            mapping[name] = dumps(value)

    def delete(self, name: str, map_name: typing.Optional[str] = None) -> None:
        with self._in_queue():
            used_map = self.key_map
            if map_name is not None:
                if map_name not in self.key_map:
                    return
                used_map = self.key_map[map_name]
                if isinstance(used_map, tuple):
                    return

            with suppress(KeyError):
                used_map.pop(name, None)



from enum import Enum


class Driver(str, Enum):
    """Enum representing database drivers.

    Attributes:
        MYSQL (str): MySQL driver.
        POSTGRES (str): PostgreSQL driver.
        SQLITE (str): SQLite driver.
        MARIADB (str): MariaDB driver.
        CUSTOM (str): Custom driver.
    """

    MYSQL = "mysql"
    POSTGRES = "postgres"
    SQLITE = "sqlite"
    MARIADB = "mariadb"
    CUSTOM = "custom"


class OrderDirection(str, Enum):
    """Enum representing order directions.

    Attributes:
        ASC (str): Ascending order.
        DESC (str): Descending order.
        DISABLED (str): Disabled order.
    """

    ASC = "asc"
    DESC = "desc"
    DISABLED = "disabled"


class ClauseType(str, Enum):
    """Enum representing clause types.

    Attributes:
        BIND (str): Bind clause type.
        APPLY (str): Apply clause type.
    """

    BIND = "bind"
    APPLY = "apply"



from typing import Any
from typing import TypeVar
from typing import final

import sqlalchemy as sa
from sqlalchemy.orm import as_declarative
from sqlalchemy.orm import declared_attr

from .metadata import metadata

T = TypeVar("T")


def _wrap(cls: type[T]) -> type[T]:
    return as_declarative(metadata=metadata)(cls)


@_wrap
class AbstractEntity:
    """Default Entity Representation without default primary key"""

    @final
    @declared_attr  # type:ignore
    def __tablename__(self):
        return self.classname()

    @classmethod
    def classname(cls):
        """Override this method to change tablename default"""
        return cls.__name__.lower()


class Entity(AbstractEntity):
    """Entity Representation with default id autoincrement as primary key"""

    __abstract__ = True

    @final
    @declared_attr
    def id_(self):
        return sa.Column("id", sa.Integer, primary_key=True)

    @property
    def pk(self):
        return self.id_


def make_table(name: str, *args: Any, **kwargs: Any) -> sa.Table:
    return sa.Table(name, metadata, *args, **kwargs)



from typing import Union
from typing import overload

from gyver.context.context import AsyncContext
from gyver.context.context import Context
from gyver.context.typedef import T

from .bound import AsyncBoundContext
from .bound import BoundContext
from .core import AsyncAtomicContext
from .core import AtomicContext


@overload
def atomic(context: Context[T], bound: bool = True) -> AtomicContext[T]:
    """
    Create an atomic context or a bound context based on the given context.

    Args:
        context (Context[T]): The base context to be wrapped.
        bound (bool, optional): Whether to create a bound context. Defaults to True.

    Returns:
        AtomicContext[T]: The created atomic or bound context.

    Raises:
        ValueError: If the provider received is not compliant with the atomic adapter interface.
    """
    ...


@overload
def atomic(context: AsyncContext[T], bound: bool = True) -> AsyncAtomicContext[T]:
    """
    Create an asynchronous atomic context or a bound context based on the given context.

    Args:
        context (AsyncContext[T]): The base asynchronous context to be wrapped.
        bound (bool, optional): Whether to create a bound context. Defaults to True.

    Returns:
        AsyncAtomicContext[T]: The created asynchronous atomic or bound context.

    Raises:
        ValueError: If the provider received is not compliant with the atomic adapter interface.
    """
    ...


def atomic(
    context: Union[Context[T], AsyncContext[T]], bound: bool = True
) -> Union[Context[T], AsyncContext[T]]:
    """
    Create an atomic context or a bound context based on the given context.

    Args:
        context (Union[Context[T], AsyncContext[T]]): The base context to be wrapped.
        bound (bool, optional): Whether to create a bound context. Defaults to True.

    Returns:
        Union[Context[T], AsyncContext[T]]: The created atomic or bound context.

    Raises:
        ValueError: If the provider received is not compliant with the atomic adapter interface.
    """
    required_methods = {"begin", "commit", "rollback", "in_atomic"}
    adapter_methods = set(dir(context.adapter))

    if not required_methods.issubset(adapter_methods):
        raise ValueError(
            f"The adapter does not have required methods {required_methods}"
        )

    if bound:
        return (
            BoundContext(context.adapter, context)  # type: ignore
            if isinstance(context, Context)
            else AsyncBoundContext(context.adapter, context)  # type: ignore
        )

    return (
        AtomicContext(context.adapter)  # type: ignore
        if isinstance(context, Context)
        else AsyncAtomicContext(context.adapter)  # type: ignore
    )



import functools
import warnings
from collections import deque
from itertools import chain
from typing import Callable
from typing import Literal
from typing import TypeVar
from typing import cast

from gyver.attrs import define
from typing_extensions import ParamSpec

from gyver.exc import MergeConflict

P = ParamSpec("P")
T = TypeVar("T")


def cache(f: Callable[P, T]) -> Callable[P, T]:
    """
    Cache the result of a function.
    It uses `functools.cache` and just
    adds proper type hints to it.

    Args:
        f (Callable[P, T]): The function to be cached.

    Returns:
        Callable[P, T]: The cached version of the function.
    """
    return cast(Callable[P, T], functools.cache(f))


def deprecated(func: Callable[P, T]) -> Callable[P, T]:
    """
    Mark a function as deprecated and issue a warning when it's used.

    Args:
        func (Callable[P, T]): The function to be marked as deprecated.

    Returns:
        Callable[P, T]: A wrapped version of the function that issues a warning on use.
    """

    @functools.wraps(func)
    def inner(*args: P.args, **kwargs: P.kwargs) -> T:
        if not hasattr(func, "__warn_deprecated__"):
            warnings.warn(
                f"{func.__qualname__} is "
                "deprecated and can be removed without notice",
                DeprecationWarning,
            )
            func.__warn_deprecated__ = True
        return func(*args, **kwargs)

    return inner


frozen = deprecated(define)


class DeprecatedClass:
    """
    A class to mark as deprecated.

    This class issues a deprecation warning when instantiated.
    """

    __warn_deprecated__ = False

    def __init__(self) -> None:
        if not type(self).__warn_deprecated__:
            warnings.warn(
                f"{'.'.join((type(self).__module__, type(self).__name__))} is "
                "deprecated and can be removed without notice",
                DeprecationWarning,
            )
            type(self).__warn_deprecated__ = True


def merge_dicts(
    left: dict,
    right: dict,
    on_conflict: Literal["strict", "left", "right"],
    merge_sequences: bool = True,
) -> dict:
    """
    Merge two dictionaries with customizable conflict resolution strategy.

    Args:
        left (dict): The left dictionary to merge.
        right (dict): The right dictionary to merge.
        on_conflict (Literal["strict", "left", "right"]): The conflict resolution strategy to use.

            - 'strict': Raise a MergeConflict exception if conflicts occur.
            - 'left': Prioritize the values from the left dictionary in case of conflicts.
            - 'right': Prioritize the values from the right dictionary in case of conflicts.
        merge_sequences (bool, optional): Indicates whether to merge sequences (lists, sets, tuples) or skip them.

            - If True, sequences will be merged based on the conflict resolution strategy.
            - If False, sequences will be skipped, and the value from the chosen (defaults to left on strict)
            dictionary will be used. Default is True.

    Returns:
        dict: The merged dictionary.

    Raises:
        MergeConflict: If conflicts occur and the conflict resolution strategy is set to 'strict'.
    """

    output = {key: value for key, value in left.items() if key not in right}

    stack = deque([(left, right, output)])

    while stack:
        left_curr, right_curr, output_curr = stack.pop()

        for key, value in right_curr.items():
            if key not in left_curr:
                output_curr[key] = value
            elif isinstance(value, (list, set, tuple)) and merge_sequences:
                left_val = left_curr[key]
                if isinstance(left_val, (list, set, tuple)):
                    type_ = type(value) if on_conflict == "right" else type(left_val)
                    output_curr[key] = type_(chain(left_val, value))
            elif isinstance(value, dict):
                if isinstance(left_curr[key], dict):
                    output_curr[key] = {
                        lkey: lvalue
                        for lkey, lvalue in left_curr[key].items()
                        if lkey not in value
                    }
                    stack.append((left_curr[key], value, output_curr[key]))
            elif on_conflict not in ("left", "right"):
                raise MergeConflict(
                    "Conflict found when trying to merge dicts",
                    key,
                    value,
                    left_curr[key],
                )
            elif on_conflict == "left":
                output_curr[key] = left_curr[key]
            else:
                output_curr[key] = value

    return output
